@@ -1,4 +1,4 @@
-/* $Id: miniupnpc.c,v 1.165 2025/01/10 22:57:21 nanard Exp $ */
+/* $Id: miniupnpc.c,v 1.166 2025/03/02 01:16:23 nanard Exp $ */
 /* vim: tabstop=4 shiftwidth=4 noexpandtab
  * Project : miniupnp
  * Web : http://miniupnp.free.fr/ or https://miniupnp.tuxfamily.org/
@@ -99,7 +99,7 @@ MINIUPNP_LIBSPEC void parserootdesc(const char * buffer, int bufsize, struct IGD
  *   NULL - error */
 char *
 simpleUPnPcommand(const char * url, const char * service,
-                  const char * action, struct UPNParg * args,
+                  const char * action, const struct UPNParg * args,
                   int * bufsize)
 {
 	char hostname[MAXHOSTNAMELEN+1];
