@@ -1,4 +1,4 @@
-/* $Id: upnpcommands.c,v 1.52 2025/01/10 22:57:21 nanard Exp $ */
+/* $Id: upnpcommands.c,v 1.54 2025/03/02 01:14:01 nanard Exp $ */
 /* vim: tabstop=4 shiftwidth=4 noexpandtab
  * Project : miniupnp
  * Author : Thomas Bernard
@@ -14,6 +14,17 @@
 #include "portlistingparse.h"
 #include "upnpreplyparse.h"
 
+/*! \file upnpcommands.c
+ * \brief Internet Gateway Device methods implementations
+ * \def STRTOUI
+ * \brief strtoull() if available, strtol() if not
+ */
+#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
+#define STRTOUI	strtoull
+#else
+#define STRTOUI	strtoul
+#endif
+
 static UNSIGNED_INTEGER
 my_atoui(const char * s)
 {
@@ -158,9 +169,7 @@ UPNP_GetStatusInfo(const char * controlURL,
 	}
 
 	if(uptime) {
-		if(up)
-			sscanf(up,"%u",uptime);
-		else
+		if(!up || sscanf(up,"%u",uptime) != 1)
 			*uptime = 0;
 	}
 
@@ -174,8 +183,8 @@ UPNP_GetStatusInfo(const char * controlURL,
 
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	ClearNameValueList(&pdata);
 	return ret;
@@ -214,8 +223,8 @@ UPNP_GetConnectionTypeInfo(const char * controlURL,
 		connectionType[0] = '\0';
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	ClearNameValueList(&pdata);
 	return ret;
@@ -262,22 +271,18 @@ UPNP_GetLinkLayerMaxBitRates(const char * controlURL,
 		ret = UPNPCOMMAND_SUCCESS;
 
 	if(bitrateDown) {
-		if(down)
-			sscanf(down,"%u",bitrateDown);
-		else
+		if(!down || sscanf(down,"%u",bitrateDown) != 1)
 			*bitrateDown = 0;
 	}
 
 	if(bitrateUp) {
-		if(up)
-			sscanf(up,"%u",bitrateUp);
-		else
+		if(!up || sscanf(up,"%u",bitrateUp) != 1)
 			*bitrateUp = 0;
 	}
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	ClearNameValueList(&pdata);
 	return ret;
@@ -327,8 +332,8 @@ UPNP_GetExternalIPAddress(const char * controlURL,
 
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 
 	ClearNameValueList(&pdata);
@@ -345,7 +350,17 @@ UPNP_AddPortMapping(const char * controlURL, const char * servicetype,
 		    const char * remoteHost,
 		    const char * leaseDuration)
 {
-	struct UPNParg * AddPortMappingArgs;
+	struct UPNParg AddPortMappingArgs[] = {
+		{"NewRemoteHost", remoteHost},
+		{"NewExternalPort", extPort},
+		{"NewProtocol", proto},
+		{"NewInternalPort", inPort},
+		{"NewInternalClient", inClient},
+		{"NewEnabled", "1"},
+		{"NewPortMappingDescription", desc?desc:"libminiupnpc"},
+		{"NewLeaseDuration", leaseDuration?leaseDuration:"0"},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	struct NameValueParserData pdata;
@@ -355,29 +370,9 @@ UPNP_AddPortMapping(const char * controlURL, const char * servicetype,
 	if(!inPort || !inClient || !proto || !extPort)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	AddPortMappingArgs = calloc(9, sizeof(struct UPNParg));
-	if(AddPortMappingArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	AddPortMappingArgs[0].elt = "NewRemoteHost";
-	AddPortMappingArgs[0].val = remoteHost;
-	AddPortMappingArgs[1].elt = "NewExternalPort";
-	AddPortMappingArgs[1].val = extPort;
-	AddPortMappingArgs[2].elt = "NewProtocol";
-	AddPortMappingArgs[2].val = proto;
-	AddPortMappingArgs[3].elt = "NewInternalPort";
-	AddPortMappingArgs[3].val = inPort;
-	AddPortMappingArgs[4].elt = "NewInternalClient";
-	AddPortMappingArgs[4].val = inClient;
-	AddPortMappingArgs[5].elt = "NewEnabled";
-	AddPortMappingArgs[5].val = "1";
-	AddPortMappingArgs[6].elt = "NewPortMappingDescription";
-	AddPortMappingArgs[6].val = desc?desc:"libminiupnpc";
-	AddPortMappingArgs[7].elt = "NewLeaseDuration";
-	AddPortMappingArgs[7].val = leaseDuration?leaseDuration:"0";
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "AddPortMapping", AddPortMappingArgs,
 	                           &bufsize);
-	free(AddPortMappingArgs);
 	if(!buffer) {
 		return UPNPCOMMAND_HTTP_ERROR;
 	}
@@ -389,8 +384,8 @@ UPNP_AddPortMapping(const char * controlURL, const char * servicetype,
 	resVal = GetValueFromNameValueList(&pdata, "errorCode");
 	if(resVal) {
 		/*printf("AddPortMapping errorCode = '%s'\n", resVal); */
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(resVal, "%d", &ret);
+		if(sscanf(resVal, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	} else {
 		ret = UPNPCOMMAND_SUCCESS;
 	}
@@ -409,7 +404,17 @@ UPNP_AddAnyPortMapping(const char * controlURL, const char * servicetype,
 		       const char * leaseDuration,
 		       char * reservedPort)
 {
-	struct UPNParg * AddPortMappingArgs;
+	struct UPNParg AddAnyPortMappingArgs[] = {
+		{"NewRemoteHost", remoteHost},
+		{"NewExternalPort", extPort},
+		{"NewProtocol", proto},
+		{"NewInternalPort", inPort},
+		{"NewInternalClient", inClient},
+		{"NewEnabled", "1"},
+		{"NewPortMappingDescription", desc?desc:"libminiupnpc"},
+		{"NewLeaseDuration", leaseDuration?leaseDuration:"0"},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	struct NameValueParserData pdata;
@@ -418,30 +423,9 @@ UPNP_AddAnyPortMapping(const char * controlURL, const char * servicetype,
 
 	if(!inPort || !inClient || !proto || !extPort)
 		return UPNPCOMMAND_INVALID_ARGS;
-
-	AddPortMappingArgs = calloc(9, sizeof(struct UPNParg));
-	if(AddPortMappingArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	AddPortMappingArgs[0].elt = "NewRemoteHost";
-	AddPortMappingArgs[0].val = remoteHost;
-	AddPortMappingArgs[1].elt = "NewExternalPort";
-	AddPortMappingArgs[1].val = extPort;
-	AddPortMappingArgs[2].elt = "NewProtocol";
-	AddPortMappingArgs[2].val = proto;
-	AddPortMappingArgs[3].elt = "NewInternalPort";
-	AddPortMappingArgs[3].val = inPort;
-	AddPortMappingArgs[4].elt = "NewInternalClient";
-	AddPortMappingArgs[4].val = inClient;
-	AddPortMappingArgs[5].elt = "NewEnabled";
-	AddPortMappingArgs[5].val = "1";
-	AddPortMappingArgs[6].elt = "NewPortMappingDescription";
-	AddPortMappingArgs[6].val = desc?desc:"libminiupnpc";
-	AddPortMappingArgs[7].elt = "NewLeaseDuration";
-	AddPortMappingArgs[7].val = leaseDuration?leaseDuration:"0";
 	buffer = simpleUPnPcommand(controlURL, servicetype,
-	                           "AddAnyPortMapping", AddPortMappingArgs,
+	                           "AddAnyPortMapping", AddAnyPortMappingArgs,
 	                           &bufsize);
-	free(AddPortMappingArgs);
 	if(!buffer) {
 		return UPNPCOMMAND_HTTP_ERROR;
 	}
@@ -449,8 +433,8 @@ UPNP_AddAnyPortMapping(const char * controlURL, const char * servicetype,
 	free(buffer);
 	resVal = GetValueFromNameValueList(&pdata, "errorCode");
 	if(resVal) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(resVal, "%d", &ret);
+		if(sscanf(resVal, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	} else {
 		char *p;
 
@@ -473,7 +457,12 @@ UPNP_DeletePortMapping(const char * controlURL, const char * servicetype,
                        const char * remoteHost)
 {
 	/*struct NameValueParserData pdata;*/
-	struct UPNParg * DeletePortMappingArgs;
+	struct UPNParg DeletePortMappingArgs[] = {
+		{"NewRemoteHost", remoteHost},
+		{"NewExternalPort", extPort},
+		{"NewProtocol", proto},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	struct NameValueParserData pdata;
@@ -483,19 +472,9 @@ UPNP_DeletePortMapping(const char * controlURL, const char * servicetype,
 	if(!extPort || !proto)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	DeletePortMappingArgs = calloc(4, sizeof(struct UPNParg));
-	if(DeletePortMappingArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	DeletePortMappingArgs[0].elt = "NewRemoteHost";
-	DeletePortMappingArgs[0].val = remoteHost;
-	DeletePortMappingArgs[1].elt = "NewExternalPort";
-	DeletePortMappingArgs[1].val = extPort;
-	DeletePortMappingArgs[2].elt = "NewProtocol";
-	DeletePortMappingArgs[2].val = proto;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                          "DeletePortMapping",
 	                          DeletePortMappingArgs, &bufsize);
-	free(DeletePortMappingArgs);
 	if(!buffer) {
 		return UPNPCOMMAND_HTTP_ERROR;
 	}
@@ -504,8 +483,8 @@ UPNP_DeletePortMapping(const char * controlURL, const char * servicetype,
 	free(buffer);
 	resVal = GetValueFromNameValueList(&pdata, "errorCode");
 	if(resVal) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(resVal, "%d", &ret);
+		if(sscanf(resVal, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	} else {
 		ret = UPNPCOMMAND_SUCCESS;
 	}
@@ -519,7 +498,13 @@ UPNP_DeletePortMappingRange(const char * controlURL, const char * servicetype,
                             const char * proto,
                             const char * manage)
 {
-	struct UPNParg * DeletePortMappingArgs;
+	struct UPNParg DeletePortMappingRangeArgs[] = {
+		{"NewStartPort", extPortStart},
+		{"NewEndPort", extPortEnd},
+		{"NewProtocol", proto},
+		{"NewManage", manage},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	struct NameValueParserData pdata;
@@ -529,22 +514,10 @@ UPNP_DeletePortMappingRange(const char * controlURL, const char * servicetype,
 	if(!extPortStart || !extPortEnd || !proto || !manage)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	DeletePortMappingArgs = calloc(5, sizeof(struct UPNParg));
-	if(DeletePortMappingArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	DeletePortMappingArgs[0].elt = "NewStartPort";
-	DeletePortMappingArgs[0].val = extPortStart;
-	DeletePortMappingArgs[1].elt = "NewEndPort";
-	DeletePortMappingArgs[1].val = extPortEnd;
-	DeletePortMappingArgs[2].elt = "NewProtocol";
-	DeletePortMappingArgs[2].val = proto;
-	DeletePortMappingArgs[3].elt = "NewManage";
-	DeletePortMappingArgs[3].val = manage;
 
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "DeletePortMappingRange",
-	                           DeletePortMappingArgs, &bufsize);
-	free(DeletePortMappingArgs);
+	                           DeletePortMappingRangeArgs, &bufsize);
 	if(!buffer) {
 		return UPNPCOMMAND_HTTP_ERROR;
 	}
@@ -552,8 +525,8 @@ UPNP_DeletePortMappingRange(const char * controlURL, const char * servicetype,
 	free(buffer);
 	resVal = GetValueFromNameValueList(&pdata, "errorCode");
 	if(resVal) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(resVal, "%d", &ret);
+		if(sscanf(resVal, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	} else {
 		ret = UPNPCOMMAND_SUCCESS;
 	}
@@ -575,24 +548,21 @@ UPNP_GetGenericPortMappingEntry(const char * controlURL,
 							 char * duration)
 {
 	struct NameValueParserData pdata;
-	struct UPNParg * GetPortMappingArgs;
+	struct UPNParg GetPortMappingArgs[] = {
+		{"NewPortMappingIndex", index},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	char * p;
-	int r = UPNPCOMMAND_UNKNOWN_ERROR;
+	int ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	if(!index)
 		return UPNPCOMMAND_INVALID_ARGS;
 	intClient[0] = '\0';
 	intPort[0] = '\0';
-	GetPortMappingArgs = calloc(2, sizeof(struct UPNParg));
-	if(GetPortMappingArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	GetPortMappingArgs[0].elt = "NewPortMappingIndex";
-	GetPortMappingArgs[0].val = index;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "GetGenericPortMappingEntry",
 	                           GetPortMappingArgs, &bufsize);
-	free(GetPortMappingArgs);
 	if(!buffer) {
 		return UPNPCOMMAND_HTTP_ERROR;
 	}
@@ -610,7 +580,7 @@ UPNP_GetGenericPortMappingEntry(const char * controlURL,
 	{
 		strncpy(extPort, p, 6);
 		extPort[5] = '\0';
-		r = UPNPCOMMAND_SUCCESS;
+		ret = UPNPCOMMAND_SUCCESS;
 	}
 	p = GetValueFromNameValueList(&pdata, "NewProtocol");
 	if(p && protocol)
@@ -623,7 +593,7 @@ UPNP_GetGenericPortMappingEntry(const char * controlURL,
 	{
 		strncpy(intClient, p, 16);
 		intClient[15] = '\0';
-		r = 0;
+		ret = 0;
 	}
 	p = GetValueFromNameValueList(&pdata, "NewInternalPort");
 	if(p)
@@ -651,11 +621,11 @@ UPNP_GetGenericPortMappingEntry(const char * controlURL,
 	}
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p) {
-		r = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &r);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	ClearNameValueList(&pdata);
-	return r;
+	return ret;
 }
 
 MINIUPNP_LIBSPEC int
@@ -688,8 +658,8 @@ UPNP_GetPortMappingNumberOfEntries(const char * controlURL,
 
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 
 	ClearNameValueList(&pdata);
@@ -712,7 +682,12 @@ UPNP_GetSpecificPortMappingEntry(const char * controlURL,
                                  char * leaseDuration)
 {
 	struct NameValueParserData pdata;
-	struct UPNParg * GetPortMappingArgs;
+	struct UPNParg GetPortMappingArgs[] = {
+		{"NewRemoteHost", remoteHost},
+		{"NewExternalPort", extPort},
+		{"NewProtocol", proto},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	char * p;
@@ -721,19 +696,9 @@ UPNP_GetSpecificPortMappingEntry(const char * controlURL,
 	if(!intPort || !intClient || !extPort || !proto)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	GetPortMappingArgs = calloc(4, sizeof(struct UPNParg));
-	if(GetPortMappingArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	GetPortMappingArgs[0].elt = "NewRemoteHost";
-	GetPortMappingArgs[0].val = remoteHost;
-	GetPortMappingArgs[1].elt = "NewExternalPort";
-	GetPortMappingArgs[1].val = extPort;
-	GetPortMappingArgs[2].elt = "NewProtocol";
-	GetPortMappingArgs[2].val = proto;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "GetSpecificPortMappingEntry",
 	                           GetPortMappingArgs, &bufsize);
-	free(GetPortMappingArgs);
 	if(!buffer) {
 		return UPNPCOMMAND_HTTP_ERROR;
 	}
@@ -777,8 +742,8 @@ UPNP_GetSpecificPortMappingEntry(const char * controlURL,
 
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 
 	ClearNameValueList(&pdata);
@@ -803,7 +768,14 @@ UPNP_GetListOfPortMappings(const char * controlURL,
                            struct PortMappingParserData * data)
 {
 	struct NameValueParserData pdata;
-	struct UPNParg * GetListOfPortMappingsArgs;
+	struct UPNParg GetListOfPortMappingsArgs[] = {
+		{"NewStartPort", startPort},
+		{"NewEndPort", endPort},
+		{"NewProtocol", protocol},
+		{"NewManage", "1"},
+		{"NewNumberOfPorts", numberOfPorts?numberOfPorts:"1000"},
+		{NULL, NULL}
+	};
 	const char * p;
 	char * buffer;
 	int bufsize;
@@ -812,24 +784,9 @@ UPNP_GetListOfPortMappings(const char * controlURL,
 	if(!startPort || !endPort || !protocol)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	GetListOfPortMappingsArgs = calloc(6, sizeof(struct UPNParg));
-	if(GetListOfPortMappingsArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	GetListOfPortMappingsArgs[0].elt = "NewStartPort";
-	GetListOfPortMappingsArgs[0].val = startPort;
-	GetListOfPortMappingsArgs[1].elt = "NewEndPort";
-	GetListOfPortMappingsArgs[1].val = endPort;
-	GetListOfPortMappingsArgs[2].elt = "NewProtocol";
-	GetListOfPortMappingsArgs[2].val = protocol;
-	GetListOfPortMappingsArgs[3].elt = "NewManage";
-	GetListOfPortMappingsArgs[3].val = "1";
-	GetListOfPortMappingsArgs[4].elt = "NewNumberOfPorts";
-	GetListOfPortMappingsArgs[4].val = numberOfPorts?numberOfPorts:"1000";
-
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "GetListOfPortMappings",
 	                           GetListOfPortMappingsArgs, &bufsize);
-	free(GetListOfPortMappingsArgs);
 	if(!buffer) {
 		return UPNPCOMMAND_HTTP_ERROR;
 	}
@@ -866,8 +823,8 @@ UPNP_GetListOfPortMappings(const char * controlURL,
 
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p) {
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	ClearNameValueList(&pdata);
 
@@ -914,8 +871,8 @@ UPNP_GetFirewallStatus(const char * controlURL,
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p)
 	{
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	ClearNameValueList(&pdata);
 	return ret;
@@ -930,7 +887,14 @@ UPNP_GetOutboundPinholeTimeout(const char * controlURL, const char * servicetype
                     const char * proto,
                     int * opTimeout)
 {
-	struct UPNParg * GetOutboundPinholeTimeoutArgs;
+	struct UPNParg GetOutboundPinholeTimeoutArgs[] = {
+		{"RemoteHost", remoteHost},
+		{"RemotePort", remotePort},
+		{"Protocol", proto},
+		{"InternalPort", intPort},
+		{"InternalClient", intClient},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	struct NameValueParserData pdata;
@@ -940,22 +904,8 @@ UPNP_GetOutboundPinholeTimeout(const char * controlURL, const char * servicetype
 	if(!intPort || !intClient || !proto || !remotePort || !remoteHost)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	GetOutboundPinholeTimeoutArgs = calloc(6, sizeof(struct UPNParg));
-	if(GetOutboundPinholeTimeoutArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	GetOutboundPinholeTimeoutArgs[0].elt = "RemoteHost";
-	GetOutboundPinholeTimeoutArgs[0].val = remoteHost;
-	GetOutboundPinholeTimeoutArgs[1].elt = "RemotePort";
-	GetOutboundPinholeTimeoutArgs[1].val = remotePort;
-	GetOutboundPinholeTimeoutArgs[2].elt = "Protocol";
-	GetOutboundPinholeTimeoutArgs[2].val = proto;
-	GetOutboundPinholeTimeoutArgs[3].elt = "InternalPort";
-	GetOutboundPinholeTimeoutArgs[3].val = intPort;
-	GetOutboundPinholeTimeoutArgs[4].elt = "InternalClient";
-	GetOutboundPinholeTimeoutArgs[4].val = intClient;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "GetOutboundPinholeTimeout", GetOutboundPinholeTimeoutArgs, &bufsize);
-	free(GetOutboundPinholeTimeoutArgs);
 	if(!buffer)
 		return UPNPCOMMAND_HTTP_ERROR;
 	ParseNameValue(buffer, bufsize, &pdata);
@@ -963,8 +913,8 @@ UPNP_GetOutboundPinholeTimeout(const char * controlURL, const char * servicetype
 	resVal = GetValueFromNameValueList(&pdata, "errorCode");
 	if(resVal)
 	{
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(resVal, "%d", &ret);
+		if(sscanf(resVal, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	else
 	{
@@ -987,7 +937,15 @@ UPNP_AddPinhole(const char * controlURL, const char * servicetype,
                     const char * leaseTime,
                     char * uniqueID)
 {
-	struct UPNParg * AddPinholeArgs;
+	struct UPNParg AddPinholeArgs[] = {
+		{"RemoteHost", ""},
+		{"RemotePort", remotePort},
+		{"Protocol", proto},
+		{"InternalPort", intPort},
+		{"InternalClient", ""},
+		{"LeaseTime", leaseTime},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	struct NameValueParserData pdata;
@@ -998,41 +956,14 @@ UPNP_AddPinhole(const char * controlURL, const char * servicetype,
 	if(!intPort || !intClient || !proto || !remoteHost || !remotePort || !leaseTime)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	AddPinholeArgs = calloc(7, sizeof(struct UPNParg));
-	if(AddPinholeArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	/* RemoteHost can be wilcarded */
-	if(strncmp(remoteHost, "empty", 5)==0)
-	{
-		AddPinholeArgs[0].elt = "RemoteHost";
-		AddPinholeArgs[0].val = "";
-	}
-	else
-	{
-		AddPinholeArgs[0].elt = "RemoteHost";
+	/* RemoteHost and InternalClient can be wilcarded
+	 * accept both the empty string and "empty" as wildcard */
+	if(strncmp(remoteHost, "empty", 5) != 0)
 		AddPinholeArgs[0].val = remoteHost;
-	}
-	AddPinholeArgs[1].elt = "RemotePort";
-	AddPinholeArgs[1].val = remotePort;
-	AddPinholeArgs[2].elt = "Protocol";
-	AddPinholeArgs[2].val = proto;
-	AddPinholeArgs[3].elt = "InternalPort";
-	AddPinholeArgs[3].val = intPort;
-	if(strncmp(intClient, "empty", 5)==0)
-	{
-		AddPinholeArgs[4].elt = "InternalClient";
-		AddPinholeArgs[4].val = "";
-	}
-	else
-	{
-		AddPinholeArgs[4].elt = "InternalClient";
+	if(strncmp(intClient, "empty", 5) != 0)
 		AddPinholeArgs[4].val = intClient;
-	}
-	AddPinholeArgs[5].elt = "LeaseTime";
-	AddPinholeArgs[5].val = leaseTime;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "AddPinhole", AddPinholeArgs, &bufsize);
-	free(AddPinholeArgs);
 	if(!buffer)
 		return UPNPCOMMAND_HTTP_ERROR;
 	ParseNameValue(buffer, bufsize, &pdata);
@@ -1047,8 +978,8 @@ UPNP_AddPinhole(const char * controlURL, const char * servicetype,
 	if(resVal)
 	{
 		/*printf("AddPortMapping errorCode = '%s'\n", resVal);*/
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(resVal, "%d", &ret);
+		if(sscanf(resVal, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	else
 	{
@@ -1063,7 +994,11 @@ UPNP_UpdatePinhole(const char * controlURL, const char * servicetype,
                     const char * uniqueID,
                     const char * leaseTime)
 {
-	struct UPNParg * UpdatePinholeArgs;
+	struct UPNParg UpdatePinholeArgs[] = {
+		{"UniqueID", uniqueID},
+		{"NewLeaseTime", leaseTime},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	struct NameValueParserData pdata;
@@ -1073,16 +1008,8 @@ UPNP_UpdatePinhole(const char * controlURL, const char * servicetype,
 	if(!uniqueID || !leaseTime)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	UpdatePinholeArgs = calloc(3, sizeof(struct UPNParg));
-	if(UpdatePinholeArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	UpdatePinholeArgs[0].elt = "UniqueID";
-	UpdatePinholeArgs[0].val = uniqueID;
-	UpdatePinholeArgs[1].elt = "NewLeaseTime";
-	UpdatePinholeArgs[1].val = leaseTime;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "UpdatePinhole", UpdatePinholeArgs, &bufsize);
-	free(UpdatePinholeArgs);
 	if(!buffer)
 		return UPNPCOMMAND_HTTP_ERROR;
 	ParseNameValue(buffer, bufsize, &pdata);
@@ -1091,8 +1018,8 @@ UPNP_UpdatePinhole(const char * controlURL, const char * servicetype,
 	if(resVal)
 	{
 		/*printf("AddPortMapping errorCode = '%s'\n", resVal); */
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(resVal, "%d", &ret);
+		if(sscanf(resVal, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	else
 	{
@@ -1106,7 +1033,10 @@ MINIUPNP_LIBSPEC int
 UPNP_DeletePinhole(const char * controlURL, const char * servicetype, const char * uniqueID)
 {
 	/*struct NameValueParserData pdata;*/
-	struct UPNParg * DeletePinholeArgs;
+	struct UPNParg DeletePinholeArgs[] = {
+		{"UniqueID", uniqueID},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	struct NameValueParserData pdata;
@@ -1116,14 +1046,8 @@ UPNP_DeletePinhole(const char * controlURL, const char * servicetype, const char
 	if(!uniqueID)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	DeletePinholeArgs = calloc(2, sizeof(struct UPNParg));
-	if(DeletePinholeArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	DeletePinholeArgs[0].elt = "UniqueID";
-	DeletePinholeArgs[0].val = uniqueID;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "DeletePinhole", DeletePinholeArgs, &bufsize);
-	free(DeletePinholeArgs);
 	if(!buffer)
 		return UPNPCOMMAND_HTTP_ERROR;
 	/*DisplayNameValueList(buffer, bufsize);*/
@@ -1132,8 +1056,8 @@ UPNP_DeletePinhole(const char * controlURL, const char * servicetype, const char
 	resVal = GetValueFromNameValueList(&pdata, "errorCode");
 	if(resVal)
 	{
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(resVal, "%d", &ret);
+		if(sscanf(resVal, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 	else
 	{
@@ -1148,7 +1072,10 @@ UPNP_CheckPinholeWorking(const char * controlURL, const char * servicetype,
                                  const char * uniqueID, int * isWorking)
 {
 	struct NameValueParserData pdata;
-	struct UPNParg * CheckPinholeWorkingArgs;
+	struct UPNParg CheckPinholeWorkingArgs[] = {
+		{"UniqueID", uniqueID},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	char * p;
@@ -1157,14 +1084,8 @@ UPNP_CheckPinholeWorking(const char * controlURL, const char * servicetype,
 	if(!uniqueID)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	CheckPinholeWorkingArgs = calloc(4, sizeof(struct UPNParg));
-	if(CheckPinholeWorkingArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	CheckPinholeWorkingArgs[0].elt = "UniqueID";
-	CheckPinholeWorkingArgs[0].val = uniqueID;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "CheckPinholeWorking", CheckPinholeWorkingArgs, &bufsize);
-	free(CheckPinholeWorkingArgs);
 	if(!buffer)
 	{
 		return UPNPCOMMAND_HTTP_ERROR;
@@ -1184,8 +1105,8 @@ UPNP_CheckPinholeWorking(const char * controlURL, const char * servicetype,
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p)
 	{
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 
 	ClearNameValueList(&pdata);
@@ -1197,7 +1118,10 @@ UPNP_GetPinholePackets(const char * controlURL, const char * servicetype,
                                  const char * uniqueID, int * packets)
 {
 	struct NameValueParserData pdata;
-	struct UPNParg * GetPinholePacketsArgs;
+	struct UPNParg GetPinholePacketsArgs[] = {
+		{"UniqueID", uniqueID},
+		{NULL, NULL}
+	};
 	char * buffer;
 	int bufsize;
 	char * p;
@@ -1206,14 +1130,8 @@ UPNP_GetPinholePackets(const char * controlURL, const char * servicetype,
 	if(!uniqueID)
 		return UPNPCOMMAND_INVALID_ARGS;
 
-	GetPinholePacketsArgs = calloc(4, sizeof(struct UPNParg));
-	if(GetPinholePacketsArgs == NULL)
-		return UPNPCOMMAND_MEM_ALLOC_ERROR;
-	GetPinholePacketsArgs[0].elt = "UniqueID";
-	GetPinholePacketsArgs[0].val = uniqueID;
 	buffer = simpleUPnPcommand(controlURL, servicetype,
 	                           "GetPinholePackets", GetPinholePacketsArgs, &bufsize);
-	free(GetPinholePacketsArgs);
 	if(!buffer)
 		return UPNPCOMMAND_HTTP_ERROR;
 	ParseNameValue(buffer, bufsize, &pdata);
@@ -1229,8 +1147,8 @@ UPNP_GetPinholePackets(const char * controlURL, const char * servicetype,
 	p = GetValueFromNameValueList(&pdata, "errorCode");
 	if(p)
 	{
-		ret = UPNPCOMMAND_UNKNOWN_ERROR;
-		sscanf(p, "%d", &ret);
+		if(sscanf(p, "%d", &ret) != 1)
+			ret = UPNPCOMMAND_UNKNOWN_ERROR;
 	}
 
 	ClearNameValueList(&pdata);
